# encoding: utf-8
"""
# load_python.py

Handlers for dumping and loading built-in python types.
NB: As these are for built-in types, they are critical to the functioning of
hickle.

"""


# %% IMPORTS
# Built-in imports
import warnings
import re
import operator

# Package imports
import numpy as np
import pickle
import h5py
import warnings

# hickle imports
from hickle.helpers import PyContainer,NotHicklable,no_compression,nobody_is_my_name


# Define conversion dict of all dict key types which shall be mapped
# to their string representation. Any other key types are directly handled
# by create_dictlike_dataset function
dict_key_types_dict = {
    b'float': float,
    b'bool': bool,
    b'int': int,
    b'complex': complex,
    b'NoneType': lambda x : None,
    # str type keys required for h5py >= 3.x
    'float': float,
    'bool': bool,
    'int': int,
    'complex': complex,
    'NoneType': lambda x : None
}

# %% FUNCTION DEFINITIONS

def create_scalar_dataset(py_obj, h_group, name, **kwargs):
    """ dumps a python dtype object to h5py file

    Parameters
    ----------
    py_obj (object):
        python object to dump; should be a scalar (int, float, bool, str, etc)

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing corresponding h5py.Dataset and empty subitems list
    """

    # If py_obj is an integer and cannot be stored in 64-bits, convert to str
    # the sign is not counted by bit_length thus any integer which has more than
    # 63 bits has to be converted into string
    if isinstance(py_obj, int) and (py_obj.bit_length() > 63):# and ( py_obj < -2**63 or py_obj >= 2**63 ) :
        return h_group.create_dataset(name,data = bytearray(str(py_obj), 'ascii'),**kwargs),()

    return h_group.create_dataset(name, data=py_obj, **no_compression(kwargs)),()


def create_none_dataset(py_obj, h_group, name, **kwargs):
    """ Dump None type to file

    Parameters
    ----------
    py_obj (NoneType):
        python object to dump; must be None object

    h_group (h5.File.Group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function
            
    Returns
    -------
    tuple containing corresponding empty h5py.Dataset and empty subitems list
    """
    return h_group.create_dataset(name, shape = None,dtype = 'V1',**no_compression(kwargs)),()


def check_iterable_item_type(first_item,iter_obj):
    """
    checks if for all items of an iterable sequence (list, tuple, etc.) a least common
    dtype exists to which all items can be safely be casted.

    Parameters
    ----------
    first_item:
        the first item of the iterable sequence used to initialize the dtype

    iter_obj:
        the remaining items of the iterable sequence

    Returns
    -------
    the least common dtype or none if not all items can be casted
    """

    if (
        operator.length_hint(first_item) > 1 or
        ( operator.length_hint(first_item) == 1 and not isinstance(first_item,(str,bytes)) ) or
        np.ndim(first_item) != 0
    ):
        return None

    dtype = np.dtype(first_item.__class__)
    if dtype.name == 'object' or 'str' in dtype.name or ( 'bytes' in dtype.name and len(first_item) > 1):
        return None
    for item in iter_obj:
        if np.ndim(item) != 0:
            return None
        common_dtype = np.result_type(np.dtype(item.__class__),dtype)
        if ( 
            common_dtype.name == 'object' or
            'str' in common_dtype.name or
            ( 'bytes' in common_dtype.name and len(item) > 1 )
        ):
            return None
        if dtype != common_dtype:
            dtype = common_dtype
    return dtype

def create_listlike_dataset(py_obj, h_group, name,list_len = -1,item_dtype = None, **kwargs):
    """ Dumper for list, set, tuple

    Parameters
    ----------
    py_obj (list, set, tuple, ...):
        python object to dump; should be list-like

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Group or h5py.Dataset representing list-like object and
    a list of subitems to be stored within this group. In case of h5py.Dataset returned
    this list is always empty 
    """

    if isinstance(py_obj,(str,bytes)):

        # strings and bytes are stored as array of bytes with strings encoded
        # using utf8 encoding
        string_data = bytearray(py_obj,"utf8") if isinstance(py_obj,str) else memoryview(py_obj)
        string_data = np.array(string_data,copy=False)
        string_data.dtype = 'S1'
        dataset = h_group.create_dataset( name, data = string_data,shape = (1,string_data.size), **kwargs)
        dataset.attrs["str_type"] = py_obj.__class__.__name__.encode("ascii")
        return dataset,()
        
    if len(py_obj) < 1:
        # list-like object is empty just store empty dataset
        return h_group.create_dataset(name,shape=None,dtype='int',**no_compression(kwargs)),()

    if list_len < 0:
        # neither length nor dtype of items is known compute them now
        item_dtype = check_iterable_item_type(py_obj[0],py_obj[1:])
        list_len = len(py_obj)

    if item_dtype or list_len < 1:
        # create a dataset and map all items to least common dtype
        shape = (list_len,) if list_len > 0 else None
        dataset = h_group.create_dataset(name,shape = shape,dtype = item_dtype,**kwargs)
        for index,item in enumerate(py_obj,0):
            dataset[index] = item_dtype.type(item)
        return dataset,()

    # create group and provide generator yielding all subitems to be stored within
    item_name = "data{:d}"
    def provide_listlike_items():
        for index,item in enumerate(py_obj,0):
            yield item_name.format(index),item,{"item_index":index},kwargs

    h_subgroup = h_group.create_group(name)
    h_subgroup.attrs["num_items"] = list_len
    return h_subgroup,provide_listlike_items()
                

def create_setlike_dataset(py_obj,h_group,name,**kwargs):
    """
    Creates a dataset or group for set-like objects. 

    Parameters
    ----------
    py_obj (set, ...):
        python object to dump; should be set-like

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Group or h5py.Dataset representing set-like object and
    a list of subitems to be stored within this group. In case of h5py.Dataset this
    list is always empty 
    """

    # set objects do not support indexing thus determination of item dtype has to
    # be handled specially. Call create_listlike_dataset for proper creation
    # of corresponding dataset
    if not py_obj:
        # dump empty set
        return h_group.create_dataset(
            name, data = list(py_obj), shape = None, dtype = int, **no_compression(kwargs)
        ),()
    set_iter = iter(py_obj)
    first_item = next(set_iter)
    item_dtype = check_iterable_item_type(first_item,set_iter)
    return create_listlike_dataset(
        py_obj, h_group, name, list_len = len(py_obj), item_dtype = item_dtype, **kwargs
    )
    

_byte_slashes = re.compile(b'[\\/]')
_str_slashes = re.compile(r'[\\/]')

def create_dictlike_dataset(py_obj, h_group, name, **kwargs):

    """ Creates a data group for each key in dictionary

    Parameters
    ----------
    py_obj (dict):
        python object to dump; should be dictionary

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        h5 node name 

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Group or h5py.Dataset representing dict-like object and
    a list of subitems to be stored within this group. In case of h5py.Dataset this
    list is always empty 
    """

    h_dictgroup = h_group.create_group(name)
    key_value_pair_name = "data{:d}"

    def package_dict_items():
        """
        generator yielding appropriate parameters for dumping each
        dict key value pair
        """
        for idx, (key, py_subobj) in enumerate(py_obj.items()):
            # Obtain the raw string representation of this key
            key_base_type = key.__class__.__name__.encode("utf8")
            if isinstance(key,str):
                if not _str_slashes.search(key):
                    yield r'"{}"'.format(key),py_subobj,{'key_idx':idx,'key_base_type':key_base_type},kwargs
                    continue
            elif isinstance(key,bytes):
                if not _byte_slashes.search(key):
                    try:
                        h_key = key.decode("utf8")
                    except UnicodeError: # pragma no cover
                        pass
                    else:
                        yield r'b"{}"'.format(h_key),py_subobj,{'key_idx':idx,'key_base_type':key_base_type},kwargs
                        continue
            elif key_base_type in dict_key_types_dict:
                h_key = "{!r}".format(key)
                if not _str_slashes.search(h_key):
                    yield h_key,py_subobj,{'key_idx':idx,'key_base_type':key_base_type},kwargs
                    continue
            sub_node_name = key_value_pair_name.format(idx)
            yield sub_node_name,(key,py_subobj),{'key_idx':idx,'key_base_type':b'key_value'},kwargs
    return h_dictgroup,package_dict_items()



def load_scalar_dataset(h_node, base_type, py_obj_type):
    """
    loads scalar dataset

    Parameters
    ----------
        h_node (h5py.Dataset):
            the hdf5 node to load data from

        base_type (bytes):
            bytes string denoting base_type

        py_obj_type (type):
            final type of restored scalar

    Returns
    -------
    resulting python object of type py_obj_type
    """
    data = h_node[()] if h_node.size < 2 else memoryview(h_node[()])


    return py_obj_type(data) if data.__class__ is not py_obj_type else data

def load_none_dataset(h_node,base_type,py_obj_type):
    """
    returns None value as represented by underlying empty dataset

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (NoneType):
        final type of restored scalar

    Returns
    -------
    None
    """
    return None
    
def load_list_dataset(h_node,base_type,py_obj_type):
    """
    loads any kind of list like dataset

    Parameters
    ----
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (list, tuple, set, ...): final type of restored object

    Returns
    -------
    resulting python object of type py_obj_type
    """

    if h_node.shape is None:

        # empty list tuple or set just return new instance of py_obj_type
        return py_obj_type() if isinstance(py_obj_type,tuple) else py_obj_type(())

    str_type = h_node.attrs.get('str_type', None)
    content = h_node[()]
    if str_type in (b'str','str'):

        # decode bytes representing python string before final conversion
        if h_node.dtype.itemsize > 1 and 'bytes' in h_node.dtype.name:

            # string dataset 4.0.x style convert it back to python string
            content = np.array(content, copy=False, dtype=str).tolist()
        else:

            # decode bytes representing python string before final conversion
            content = bytes(content).decode("utf8")
    return py_obj_type(content) if content.__class__ is not py_obj_type else content

def load_hickle_4_x_string(h_node,base_type,py_obj_type):
    """
    loads dataset representing python string stored by hickle 4.x 

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (str): final type of restored string

    Returns
    -------
    resulting python str or bytes object as specified by py_obj_type
    """
    if not 'object' in h_node.dtype.name or h_node.attrs.get('str_type',None) is not None:
        return load_list_dataset(h_node,base_type,py_obj_type)
    content = h_node[()]
    if py_obj_type is str:
        return content if isinstance(content,str) else content.decode('utf8')
    return py_obj_type(content) if content.__class__ is not py_obj_type else content
        

class ListLikeContainer(PyContainer):
    """
    PyContainer for all list like objects exempt set
    """

    __slots__ = ()

    # regular expression used to extract index value from name of group or dataset
    # representing subitem appended to the final list
    extract_index = re.compile(r'\d+$')

    # as None can and may be a valid list entry define an alternative marker for
    # missing items and indices
    def __init__(self,h5_attrs,base_type,object_type):
        # if number of items is defined upon group resize content to
        # at least match this amount of subitems
        num_items = h5_attrs.get('num_items',0)
        super(ListLikeContainer,self).__init__(h5_attrs,base_type,object_type,_content = [nobody_is_my_name] * num_items)

    def append(self,name,item,h5_attrs):
        # load item index from attributes if known else extract it from name 
        index = h5_attrs.get("item_index",None)
        if index is None:
            index_match = self.extract_index.search(name)
            if index_match is None:
                if item is nobody_is_my_name:
                    # dummy data injected likely by load_nothing, ignore it
                    return
                raise KeyError("List like item name '{}' not understood".format(name))
            index = int(index_match.group(0))
        # if index exceeds capacity of extend list appropriately
        if len(self._content) <= index:
            self._content.extend([nobody_is_my_name] * ( index - len(self._content) + 1 ))
        if self._content[index] is not nobody_is_my_name:
            raise IndexError("Index {} already set".format(index))
        self._content[index] = item
        
    def convert(self):
        return self._content if self.object_type is self._content.__class__ else self.object_type(self._content)

class SetLikeContainer(PyContainer):
    """
    PyContainer for all set like objects.
    """
    __slots__ = ()

    def __init__(self,h5_attrs, base_type, object_type):
        super(SetLikeContainer,self).__init__(h5_attrs,base_type,object_type,_content=set())

            
    def append(self,name,item,h5_attrs):
        self._content.add(item)

    def convert(self):
        return self._content if self._content.__class__ is self.object_type else self.object_type(self._content)

class DictLikeContainer(PyContainer):
    """
    PyContainer for all dict like objects
    """

    __slots__ = ()

    
    _swap_key_slashes = re.compile(r"\\")

    def append(self,name,item,h5_attrs):
        key_base_type = h5_attrs.get('key_base_type',b'')
        if key_base_type in ( b'str','str'):
            item = (
                name[1:-1] if name[0] == '"' else self._swap_key_slashes.sub(r'/',name)[1:-1],
                item
            )
        elif key_base_type in (b'bytes','bytes'):
            item = (
                name[2:-1].encode("utf8") if name[:2] == 'b"' else self._swap_key_slashes.sub(r'/',name)[1:-1],
                item
            )
        elif not key_base_type in (b'key_value','key_value'):
            load_key  = dict_key_types_dict.get(key_base_type,None)
            if load_key is None:
                if key_base_type not in {b'tuple','tuple'}:
                    raise ValueError("key type '{}' not understood".format(key_base_type.decode("utf8")))
                import ast
                load_key = ast.literal_eval
            item = (
                load_key(self._swap_key_slashes.sub(r'/',name)),
                item
            )
        key_index = h5_attrs.get('key_idx',None)
        if key_index is None:
            if item[1] is nobody_is_my_name:
                # dummy data injected most likely by load_nothing ignore it
                return
            raise KeyError("invalid dict item key_index missing")
        if len(self._content) <= key_index:
            self._content.extend([nobody_is_my_name] * ( key_index - len(self._content) + 1))
        if self._content[key_index] is not nobody_is_my_name:
            raise IndexError("Key index {} already set".format(key_index))
        self._content[key_index] = item

    def convert(self):
        return self.object_type(self._content)

    

# %% REGISTERS
class_register = [
    [list, b"list", create_listlike_dataset, load_list_dataset,ListLikeContainer],
    [tuple, b"tuple", create_listlike_dataset, load_list_dataset,ListLikeContainer],
    [dict, b"dict",create_dictlike_dataset,None,DictLikeContainer],
    [set, b"set", create_setlike_dataset, load_list_dataset,SetLikeContainer],
    [bytes, b"bytes", create_listlike_dataset, load_list_dataset],
    [str, b"str", create_listlike_dataset, load_list_dataset],
    [str, b"str", None, load_hickle_4_x_string,None,True,'hickle-4.x'],
    [int, b"int", create_scalar_dataset, load_scalar_dataset, None, False],
    [float, b"float", create_scalar_dataset, load_scalar_dataset, None, False],
    [complex, b"complex", create_scalar_dataset, load_scalar_dataset, None, False],
    [bool, b"bool", create_scalar_dataset, load_scalar_dataset, None, False],
    [None.__class__, b"None", create_none_dataset, load_none_dataset, None, False]
]

exclude_register = []



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle_helpers.py

Unit tests for hickle module -- helper functions.

"""

import pytest

# %% IMPORTS
# Package imports
import numpy as np
import pickle
import operator
import numpy as np
import h5py

# hickle imports
from hickle.helpers import (
    PyContainer,H5NodeFilterProxy,no_compression,convert_str_attr,convert_str_list_attr
)
from hickle.fileio import FileError,ClosedFileError,file_opener,not_io_base_like
from py.path import local

# Set current working directory to the temporary directory
local.get_temproot().chdir()


# %% DATA DEFINITIONS

dummy_data = (1,2,3)


# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """

    # create file and create a dataset the attributes of which will later on be
    # modified
    import h5py as h5
    dummy_file = h5.File('hickle_helpers_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_dataset("somedata",data=dummy_data,dtype='i')
    test_data.attrs['type'] = np.array(pickle.dumps(tuple))
    test_data.attrs['base_type'] = b'tuple'
    test_data.attrs['someattr'] = 12
    test_data.attrs['someother'] = 11

    # write out the file reopen it read only
    dummy_file.flush()
    dummy_file.close()
    dummy_file = h5.File(filename,'r')

    # provide the file and close afterwards
    yield dummy_file
    dummy_file.close()
    
@pytest.fixture
def test_file_name(request):
    yield "{}.hkl".format(request.function.__name__)
    
# %% FUNCTION DEFINITIONS

def test_no_compression():
    """
    test no_compression filter for temporarily hiding compression related
    kwargs from h5py.create_dataset method
    """

    # simulate kwargs without compression related
    kwargs = {'hello':1,'word':2}
    assert dict(no_compression(kwargs)) == kwargs
    
    # simulate kwargs including all relevant keyword arguments
    kwargs2 = dict(kwargs)
    kwargs2.update({
        "compression":True,
        "shuffle":True,
        "compression_opts":8,
        "chunks":512,
        "fletcher32":True,
        "scaleoffset":20
    })
    assert dict(no_compression(kwargs2)) == kwargs
    
def test_py_container(h5_data):
    """
    test abstract PyContainer base class defining container interface
    and providing default implementations for append and filter
    """

    # test default implementation of append
    container = PyContainer({},b'list',list)
    container.append('data0',1,{})
    container.append('data1','b',{})

    # ensure that default implementation of convert enforces overload by
    # derived PyContainer classes by raising NotImplementedError
    with pytest.raises(NotImplementedError):
        my_list = container.convert()

    # test default implementation of PyContainer.filter method which
    # simply shall yield from passed in iterator
    assert [ item for item in dummy_data ] == list(dummy_data)
    assert dict(container.filter(h5_data)) == {'somedata':h5_data['somedata']}


def test_H5NodeFilterProxy(h5_data):
    """
    tests H5NodeFilterProxy class. This class allows to temporarily rewrite
    attributes of h5py.Group and h5py.Dataset nodes before being loaded by 
    hickle._load method. 
    """

    # load data and try to directly modify 'type' and 'base_type' Attributes
    # which will fail cause hdf5 file is opened for read only
    h5_node = h5_data['somedata']
    with pytest.raises(OSError):
        try:
            h5_node.attrs['type'] = pickle.dumps(list)
        except RuntimeError as re:
            raise OSError(re).with_traceback(re.__traceback__)
    with pytest.raises(OSError):
        try:
            h5_node.attrs['base_type'] = b'list'
        except RuntimeError as re:
            raise OSError(re).with_traceback(re.__traceback__)

    # verify that 'type' expands to tuple before running
    # the remaining tests
    object_type = pickle.loads(h5_node.attrs['type'])
    assert object_type is tuple
    assert object_type(h5_node[()].tolist()) == dummy_data

    # Wrap node by H5NodeFilterProxy and rerun the above tests
    # again. This time modifying Attributes shall be possible.
    h5_node = H5NodeFilterProxy(h5_node)
    h5_node.attrs['type'] = pickle.dumps(list)
    h5_node.attrs['base_type'] = b'list'
    object_type = pickle.loads(h5_node.attrs['type'])
    assert object_type is list

    # test proper pass through of item and attribute access 
    # to wrapped h5py.Group or h5py.Dataset object respective
    assert object_type(h5_node[()].tolist()) == list(dummy_data)
    assert h5_node.shape == np.array(dummy_data).shape
    with pytest.raises(AttributeError,match = r"can't\s+set\s+attribute"):
        h5_node.dtype = np.float32

def test_not_io_base_like(test_file_name):
    """
    test not_io_base_like function for creating replacement methods
    for IOBase.isreadable, IOBase.isseekable and IOBase.writeable
    """
    with open(test_file_name,'w') as f:
        assert not not_io_base_like(f)()
        assert not not_io_base_like(f,'strange_read',0)()
        assert not not_io_base_like(f,'seek',0,'strange_tell')()
    with open(test_file_name,'r') as f:
        assert not_io_base_like(f,('seek',0),('tell',))()
        assert not not_io_base_like(f,('seek',0),('tell',()))()
        assert not_io_base_like(f,('read',0))()
        assert not not_io_base_like(f,('tell',()))()
        assert not_io_base_like(f,('tell',))()
    

def test_file_opener(h5_data,test_file_name):
    """
    test file opener function
    """

    # check that file like object is properly initialized for writing
    filename = test_file_name.replace(".hkl","_{}.{}")
    with open(filename.format("w","hdf5"),"w") as f:
        with pytest.raises(FileError):
            h5_file,path,close_flag = file_opener(f,"root","w",filename="filename")
    with open(filename.format("w","hdf5"),"w+b") as f:
        h5_file,path,close_flag = file_opener(f,"root","w+")
        assert isinstance(h5_file,h5py.File) and path == "/root" and h5_file.mode == 'r+'
        h5_file.close()

    # check that file like object is properly initialized for reading
    with open(filename.format("w","hdf5"),"rb") as f:
        h5_file,path,close_flag = file_opener(f,"root","r")
        assert isinstance(h5_file,h5py.File) and path == "/root" and h5_file.mode == 'r'
        assert close_flag
        h5_file.close()
        # check that only str are accepted as filenames
        with pytest.raises(ValueError):
            h5_file,path,close_flag = file_opener(f,"root","r",filename=12)
        # check that tuple specifying file object and filename string is accepted
        h5_file,path,close_flag = file_opener((f,"not me"),"root","r")
        assert isinstance(h5_file,h5py.File) and path == "/root" and h5_file.mode == 'r'
        assert close_flag
        h5_file.close()
        # check that dict specifying file object and filename is accepted
        h5_file,path,close_flag = file_opener({"file":f,"name":"not me"},"root","r")
        assert isinstance(h5_file,h5py.File) and path == "/root" and h5_file.mode == 'r'
        assert close_flag
        h5_file.close()
        # check that file is rejected if mode used to open and mode passed to file
        # opener do not match
        with pytest.raises(FileError):
            h5_file,path,close_flag = file_opener({"file":f,"name":"not me"},"root","r+")
        with pytest.raises(FileError):
            h5_file,path,close_flag = file_opener({"file":f,"name":"not me"},"root","w")
        with pytest.raises(ValueError):
            h5_file,path,close_flag = file_opener({"file":f,"name":"not me"},"root","+")
    # check that only binary files opened for reading and writing are accepted with
    # mode w
    with open(filename.format("w","hdf5"),"w") as f:
        with pytest.raises(FileError):
            h5_file,path,close_flag = file_opener({"file":f,"name":"not me"},"root","w")

    # check that closed file objects are rejected
    with pytest.raises(ClosedFileError):
        h5_file,path,close_flag = file_opener(f,"root","r")
        

    # check that h5py.File object is properly initialised for writing
    with pytest.raises(FileError):
        h5_file,path,close_flag = file_opener(h5_data,"","w")
    with h5py.File(filename.format("w","whkl"),"w") as hdf5_file:
        h5_file,path,close_flag = file_opener(hdf5_file,"","w")
        assert isinstance(h5_file,h5py.File) and path == "/"
        assert h5_file.mode == 'r+' and not close_flag
        hdf5_group = hdf5_file.create_group("some_group")
    with pytest.raises(ClosedFileError):
        h5_file,path,close_flag = file_opener(hdf5_file,"","w")
    with h5py.File(filename.format("w","whkl"),"r") as hdf5_file:
        h5_file,path,close_flag = file_opener(hdf5_file["some_group"],'',"r")
        assert isinstance(h5_file,h5py.File) and path == "/some_group"
        assert h5_file.mode == 'r' and not close_flag
        

    # check that a new file is created for provided filename and properly initialized
    h5_file,path,close_flag = file_opener(filename.format("w",".hkl"),"root_group","w")
    assert isinstance(h5_file,h5py.File) and path == "/root_group"
    assert h5_file.mode == 'r+' and close_flag
    h5_file.close()
    
    # check that any other object not being a file like object, a h5py.File object or
    # a filename string triggers an  FileError exception
    with pytest.raises(FileError):
        h5_file,path,close_flag = file_opener(object(),"root_group","w")

def test_str_attr_converter():
    """
    test attribute decoder helper functions used to mimic
    h5py >= 3.x behaviour when h5py 2.10 is installed
    """
    ascii_str_val = 'some ascii encoded string attr'
    utf8_str_val = 'some utf8 encoded string attr'
    some_attrs = dict(
        some_attr_ascii = ascii_str_val.encode('ascii'),
        some_attr_utf8 = utf8_str_val.encode('utf8'),
        some_attr_list_ascii = [ strval.encode('ascii') for strval in ascii_str_val.split(' ') ],
        some_attr_list_utf8 = [ strval.encode('utf8') for strval in utf8_str_val.split(' ') ]
    )
    assert convert_str_attr(some_attrs,'some_attr_ascii',encoding='ascii') == ascii_str_val
    assert convert_str_attr(some_attrs,'some_attr_utf8') == utf8_str_val
    assert " ".join(convert_str_list_attr(some_attrs,'some_attr_list_ascii',encoding='ascii')) == ascii_str_val
    assert " ".join(convert_str_list_attr(some_attrs,'some_attr_list_utf8')) == utf8_str_val
    

# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest

    test_no_compression()
    for data in h5_data(FixtureRequest(test_py_container)):
        test_py_container(data)
    for data in h5_data(FixtureRequest(test_py_container)):
        test_H5NodeFilterProxy(data)
    for filename in (
        ( test_file_name(request), )
        for request in (FixtureRequest(test_not_io_base_like),)
    ):
        test_not_io_base_like(filename)
    for h5_root,filename in (
        ( h5_data(request),test_file_name(request) )
        for request in (FixtureRequest(test_file_opener),)
    ):
        test_file_opener(h5_root,filename)
    test_str_attr_converter()




#! /usr/bin/env python
# encoding: utf-8
"""
# test_load_scipy

Unit tests for hickle module -- scipy loader.

"""
# %% IMPORTS
# Package imports
import pytest
import h5py as h5
import numpy as np
import pickle
from scipy.sparse import csr_matrix, csc_matrix, bsr_matrix
from py.path import local

# %% HICKLE imports
import hickle.loaders.load_scipy as load_scipy

# Set the current working directory to the temporary directory
local.get_temproot().chdir()

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """
    dummy_file = h5.File('test_load_builtins.hdf5','w')
    dummy_file = h5.File('load_numpy_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()

# %% FUNCTION DEFINITIONS

def test_return_first_function_type():
    with pytest.raises(TypeError):
        load_scipy.return_first(['anything','some other thins','nothing'])

def test_create_sparse_dataset(h5_data,compression_kwargs):
    """
    test creation and loading of sparse matrix 
    """

    # create all possible kinds of sparse matrix representations
    row = np.array([0, 0, 1, 2, 2, 2])
    col = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    sm1 = csr_matrix((data, (row, col)), shape=(3, 3))
    sm2 = csc_matrix((data, (row, col)), shape=(3, 3))

    indptr = np.array([0, 2, 3, 6])
    indices = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6]).repeat(4).reshape([6, 2, 2])
    sm3 = bsr_matrix((data, indices, indptr), shape=(6, 6))

    # check that csr type matrix is properly stored and loaded
    h_datagroup,subitems = load_scipy.create_sparse_dataset(sm1,h5_data,"csr_matrix",**compression_kwargs)
    assert isinstance(h_datagroup,h5.Group) and iter(subitems)
    seen_items = dict((key,False) for key in ("data",'indices','indptr','shape'))
    sparse_container = load_scipy.SparseMatrixContainer(h_datagroup.attrs,b'csr_matrix',csr_matrix)
    for name,item,attrs,kwargs in subitems:
        assert not seen_items[name]
        seen_items[name] = True
        sparse_container.append(name,item,attrs)
    reloaded = sparse_container.convert()
    assert np.all(reloaded.data == sm1.data) and reloaded.dtype == sm1.dtype and reloaded.shape == sm1.shape

    # check that csc type matrix is properly stored and loaded
    h_datagroup,subitems = load_scipy.create_sparse_dataset(sm2,h5_data,"csc_matrix",**compression_kwargs)
    assert isinstance(h_datagroup,h5.Group) and iter(subitems)
    seen_items = dict((key,False) for key in ("data",'indices','indptr','shape'))
    sparse_container = load_scipy.SparseMatrixContainer(h_datagroup.attrs,b'csc_matrix',csc_matrix)
    for name,item,attrs,kwargs in subitems:
        assert not seen_items[name]
        seen_items[name] = True
        sparse_container.append(name,item,attrs)
    reloaded = sparse_container.convert()
    assert np.all(reloaded.data == sm2.data) and reloaded.dtype == sm2.dtype and reloaded.shape == sm2.shape

    # check that bsr type matrix is properly stored and loaded
    h_datagroup,subitems = load_scipy.create_sparse_dataset(sm3,h5_data,"bsr_matrix",**compression_kwargs)
    assert isinstance(h_datagroup,h5.Group) and iter(subitems)
    seen_items = dict((key,False) for key in ("data",'indices','indptr','shape'))
    sparse_container = load_scipy.SparseMatrixContainer(h_datagroup.attrs,b'bsr_matrix',bsr_matrix)
    for name,item,attrs,kwargs in subitems:
        assert not seen_items[name]
        seen_items[name] = True
        sparse_container.append(name,item,attrs)
    reloaded = sparse_container.convert()
    assert np.all(reloaded.data == sm3.data) and reloaded.dtype == sm3.dtype and reloaded.shape == sm3.shape

    # mimic hickle version 4.0.0 format to represent crs type matrix
    h_datagroup,subitems = load_scipy.create_sparse_dataset(sm1,h5_data,"csr_matrix_filtered",**compression_kwargs)
    sparse_container = load_scipy.SparseMatrixContainer(h_datagroup.attrs,b'csr_matrix',load_scipy.return_first)
    for name,item,attrs,kwargs in subitems:
        h_dataset = h_datagroup.create_dataset(name,data=item)
        if name == "data":
            attrs["type"] = np.array(pickle.dumps(sm1.__class__))
            attrs["base_type"] = b'csr_matrix'
        h_dataset.attrs.update(attrs)

    # check that dataset representing hickle 4.0.0 representation of sparse matrix
    # is properly recognized by SparseMatrixContainer.filter method and sub items of
    # sparse matrix group are properly adjusted to be safely loaded by SparseMatrixContainer
    for name,h_dataset in sparse_container.filter(h_datagroup):
        if name == "shape":
            sparse_container.append(name,tuple(h_dataset[()]),h_dataset.attrs)
        else:
            sparse_container.append(name,np.array(h_dataset[()]),h_dataset.attrs)
    reloaded = sparse_container.convert()
    assert np.all(reloaded.data == sm1.data) and reloaded.dtype == sm1.dtype and reloaded.shape == sm1.shape

    # verify that SparseMatrixContainer.filter method ignores any items which
    # are not recognized by SparseMatrixContainer update or convert method
    h_datagroup.create_dataset("ignoreme",data=12)
    for name,h_dataset in sparse_container.filter(h_datagroup):
        if name == "shape":
            sparse_container.append(name,tuple(h_dataset[()]),h_dataset.attrs)
        else:
            sparse_container.append(name,np.array(h_dataset[()]),h_dataset.attrs)
    reloaded = sparse_container.convert()
    assert np.all(reloaded.data == sm1.data) and reloaded.dtype == sm1.dtype and reloaded.shape == sm1.shape
    
    
# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest
    from hickle.tests.conftest import compression_kwargs

    test_return_first_function_type()
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_sparse_dataset),)
    ):
        test_create_sparse_dataset(h5_root,keywords)



#! /usr/bin/env python
# encoding: utf-8
"""
# test_load_astropy

Unit tests for hickle module -- astropy loader.

"""
# %% IMPORTS
# Package imports
import h5py as h5
import numpy as np
import pytest
from astropy.units import Quantity
from astropy.time import Time
from astropy.coordinates import Angle, SkyCoord
import astropy.constants as apc
from astropy.table import Table
import numpy as np
from py.path import local

# hickle imports
import hickle.loaders.load_astropy as load_astropy


# Set the current working directory to the temporary directory
local.get_temproot().chdir()

# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer and H5NodeFilterProxy
    """
    dummy_file = h5.File('test_load_builtins.hdf5','w')
    dummy_file = h5.File('load_numpy_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    dummy_file.close()

# %% FUNCTION DEFINITIONS
def test_create_astropy_quantity(h5_data,compression_kwargs):
    """
    test proper storage and loading of astropy quantities
    """

    for index,uu in enumerate(['m^3', 'm^3 / s', 'kg/pc']):
        a = Quantity(7, unit=uu)
        h_dataset,subitems = load_astropy.create_astropy_quantity(a,h5_data,"quantity{}".format(index),**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
        a_unit_string = a.unit.to_string()
        assert h_dataset.attrs['unit'] in ( a_unit_string.encode("ascii"),a_unit_string) and h_dataset[()] == a.value
        reloaded = load_astropy.load_astropy_quantity_dataset(h_dataset,b'astropy_quantity',Quantity)
        assert reloaded == a and reloaded.unit == a.unit
        a *= a
        h_dataset,subitems = load_astropy.create_astropy_quantity(a,h5_data,"quantity_sqr{}".format(index),**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
        a_unit_string = a.unit.to_string()
        assert h_dataset.attrs['unit'] in ( a_unit_string.encode("ascii"),a_unit_string) and h_dataset[()] == a.value
        reloaded = load_astropy.load_astropy_quantity_dataset(h_dataset,b'astropy_quantity',Quantity)
        assert reloaded == a and reloaded.unit == a.unit


def test_create_astropy_constant(h5_data,compression_kwargs):

    """
    test proper storage and loading of astropy constants
    """

    h_dataset,subitems = load_astropy.create_astropy_constant(apc.G,h5_data,"apc_G",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    apc_G_unit_string = apc.G.unit.to_string()
    assert h_dataset.attrs["unit"] in (apc_G_unit_string.encode('ascii'),apc_G_unit_string)
    assert h_dataset.attrs["abbrev"] in (apc.G.abbrev.encode('ascii'),apc.G.abbrev)
    assert h_dataset.attrs["name"] in ( apc.G.name.encode('ascii'),apc.G.name)
    assert h_dataset.attrs["reference"] in ( apc.G.reference.encode('ascii'),apc.G.reference)
    assert h_dataset.attrs["uncertainty"] == apc.G.uncertainty
    reloaded = load_astropy.load_astropy_constant_dataset(h_dataset,b'astropy_constant',apc.G.__class__)
    assert reloaded == apc.G and reloaded.dtype == apc.G.dtype

    h_dataset,subitems = load_astropy.create_astropy_constant(apc.cgs.e,h5_data,"apc_cgs_e",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert h_dataset.attrs["unit"] in ( apc.cgs.e.unit.to_string().encode('ascii'),apc.cgs.e.unit)
    assert h_dataset.attrs["abbrev"] in ( apc.cgs.e.abbrev.encode('ascii'),apc.cgs.e.abbrev)
    assert h_dataset.attrs["name"] in (apc.cgs.e.name.encode('ascii'),apc.cgs.e.name)
    assert h_dataset.attrs["reference"] in ( apc.cgs.e.reference.encode('ascii'),apc.cgs.e.reference)
    assert h_dataset.attrs["uncertainty"]  == apc.cgs.e.uncertainty
    assert h_dataset.attrs["system"] in ( apc.cgs.e.system.encode('ascii'),apc.cgs.e.system )
    reloaded = load_astropy.load_astropy_constant_dataset(h_dataset,b'astropy_constant',apc.cgs.e.__class__)
    assert reloaded == apc.cgs.e and reloaded.dtype == apc.cgs.e.dtype


def test_astropy_table(h5_data,compression_kwargs):
    """
    test proper storage and loading of astropy table
    """
    t = Table([[1, 2], [3, 4]], names=('a', 'b'), meta={'name': 'test_thing'})

    h_dataset,subitems = load_astropy.create_astropy_table(t,h5_data,"astropy_table",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert (
        np.all(h_dataset.attrs['colnames'] == [ cname.encode('ascii') for cname in t.colnames]) or
        np.all(h_dataset.attrs['colnames'] == [ cname for cname in t.colnames])
    )
        
    for metakey,metavalue in t.meta.items():
        assert h_dataset.attrs[metakey] == metavalue
    assert h_dataset.dtype == t.as_array().dtype
    reloaded = load_astropy.load_astropy_table(h_dataset,b'astropy_table',t.__class__)
    assert reloaded.meta == t.meta and reloaded.dtype == t.dtype
    assert np.allclose(t['a'].astype('float32'),reloaded['a'].astype('float32'))
    assert np.allclose(t['b'].astype('float32'),reloaded['b'].astype('float32'))


def test_astropy_quantity_array(h5_data,compression_kwargs):
    """
    test proper storage and loading of array of astropy quantities 
    """
    a = Quantity([1, 2, 3], unit='m')
    h_dataset,subitems = load_astropy.create_astropy_quantity(a,h5_data,"quantity_array",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert h_dataset.attrs['unit'] in (a.unit.to_string().encode("ascii"),a.unit.to_string()) and np.all(h_dataset[()] == a.value)
    reloaded = load_astropy.load_astropy_quantity_dataset(h_dataset,b'astropy_quantity',Quantity)
    assert np.all(reloaded == a) and reloaded.unit == a.unit


def test_astropy_time_array(h5_data,compression_kwargs):
    """
    test proper storage and loading of astropy time representations
    """
    
    loop_counter = 0

    
    for times in ([58264, 58265, 58266], [[58264, 58265, 58266], [58264, 58265, 58266]]):
        t1 = Time(times, format='mjd', scale='utc')
        
        h_dataset, subitems = load_astropy.create_astropy_time(t1,h5_data, f'time_{loop_counter}',**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
        assert h_dataset.attrs['format'] in( str(t1.format).encode('ascii'),str(t1.format))
        assert h_dataset.attrs['scale'] in ( str(t1.scale).encode('ascii'),str(t1.scale))
        assert h_dataset.attrs['np_dtype'] in( t1.value.dtype.str.encode('ascii'),t1.value.dtype.str)
        reloaded = load_astropy.load_astropy_time_dataset(h_dataset,b'astropy_time',t1.__class__)
        assert reloaded.value.shape == t1.value.shape
        assert reloaded.format == t1.format
        assert reloaded.scale == t1.scale
        for index in range(len(t1)):
            assert np.allclose(reloaded.value[index], t1.value[index])
        loop_counter += 1

    t_strings = ['1999-01-01T00:00:00.123456789', '2010-01-01T00:00:00']
    
    # Check that 2D time arrays work as well (github issue #162)
    for times in (t_strings, [t_strings, t_strings]):
        t1 = Time(times, format='isot', scale='utc')

        h_dataset,subitems = load_astropy.create_astropy_time(t1,h5_data,f'time_{loop_counter}',**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
        assert h_dataset.attrs['format'] in (str(t1.format).encode('ascii'),str(t1.format))
        assert h_dataset.attrs['scale'] in (str(t1.scale).encode('ascii'),str(t1.scale))
        assert h_dataset.attrs['np_dtype'] in ( t1.value.dtype.str.encode('ascii'),t1.value.dtype.str)
        reloaded = load_astropy.load_astropy_time_dataset(h_dataset,b'astropy_time',t1.__class__)
        assert reloaded.value.shape == t1.value.shape
        assert reloaded.format == t1.format
        assert reloaded.scale == t1.scale
        for index in range(len(t1)):
             assert reloaded.value[index].tostring() == t1.value[index].tostring()
        del h_dataset.attrs['np_dtype']

        reloaded = load_astropy.load_astropy_time_dataset(h_dataset,b'astropy_time',t1.__class__)
        assert reloaded.value.shape == t1.value.shape
        assert reloaded.format == t1.format
        assert reloaded.scale == t1.scale
        for index in range(len(t1)):
            assert reloaded.value[index].tostring() == t1.value[index].tostring()
        loop_counter += 1



def test_astropy_angle(h5_data,compression_kwargs):
    """
    test proper storage of astropy angles
    """

    for index,uu in enumerate(['radian', 'degree']):
        a = Angle(1.02, unit=uu)
        h_dataset,subitems = load_astropy.create_astropy_angle(a,h5_data,"angle_{}".format(uu),**compression_kwargs)
        assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
        assert h_dataset.attrs['unit'] in( a.unit.to_string().encode('ascii'),a.unit.to_string())
        assert h_dataset[()] == a.value
        reloaded = load_astropy.load_astropy_angle_dataset(h_dataset,b'astropy_angle',a.__class__)
        assert reloaded == a and reloaded.unit == a.unit


def test_astropy_angle_array(h5_data,compression_kwargs):
    """
    test proper storage and loading of arrays of astropy angles
    """
    a = Angle([1, 2, 3], unit='degree')
    h_dataset,subitems = load_astropy.create_astropy_angle(a,h5_data,"angle_array",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert h_dataset.attrs['unit'] in (a.unit.to_string().encode('ascii'),a.unit.to_string())
    assert np.allclose(h_dataset[()] , a.value )
    reloaded = load_astropy.load_astropy_angle_dataset(h_dataset,b'astropy_angle',a.__class__)
    assert np.all(reloaded ==  a) and reloaded.unit == a.unit

def test_astropy_skycoord(h5_data,compression_kwargs):
    """
    test proper storage and loading of astropy sky coordinates
    """

    ra = Angle('1d20m', unit='degree')
    dec = Angle('33d0m0s', unit='degree')
    radec = SkyCoord(ra, dec)
    h_dataset,subitems = load_astropy.create_astropy_skycoord(radec,h5_data,"astropy_skycoord_1",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert h_dataset[()][...,0] == radec.data.lon.value
    assert h_dataset[()][...,1] == radec.data.lat.value
    assert h_dataset.attrs['lon_unit'] in ( radec.data.lon.unit.to_string().encode('ascii'),radec.data.lon.unit.to_string())
    assert h_dataset.attrs['lat_unit'] in ( radec.data.lat.unit.to_string().encode('ascii'),radec.data.lat.unit.to_string())
    reloaded = load_astropy.load_astropy_skycoord_dataset(h_dataset,b'astropy_skycoord',radec.__class__)
    assert np.allclose(reloaded.ra.value,radec.ra.value)
    assert np.allclose(reloaded.dec.value,radec.dec.value)

    ra = Angle('1d20m', unit='hourangle')
    dec = Angle('33d0m0s', unit='degree')
    radec = SkyCoord(ra, dec)
    h_dataset,subitems = load_astropy.create_astropy_skycoord(radec,h5_data,"astropy_skycoord_2",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert h_dataset[()][...,0] == radec.data.lon.value
    assert h_dataset[()][...,1] == radec.data.lat.value
    assert h_dataset.attrs['lon_unit'] in (radec.data.lon.unit.to_string().encode('ascii'),radec.data.lon.unit.to_string())
    assert h_dataset.attrs['lat_unit'] in ( radec.data.lat.unit.to_string().encode('ascii'),radec.data.lat.unit.to_string())
    reloaded = load_astropy.load_astropy_skycoord_dataset(h_dataset,b'astropy_skycoord',radec.__class__)
    assert reloaded.ra.value == radec.ra.value
    assert reloaded.dec.value == radec.dec.value

def test_astropy_skycoord_array(h5_data,compression_kwargs):
    """
    test proper storage and loading of astropy sky coordinates
    """

    ra = Angle(['1d20m', '0d21m'], unit='degree')
    dec = Angle(['33d0m0s', '-33d01m'], unit='degree')
    radec = SkyCoord(ra, dec)
    h_dataset,subitems = load_astropy.create_astropy_skycoord(radec,h5_data,"astropy_skycoord_1",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert np.allclose(h_dataset[()][...,0],radec.data.lon.value)
    assert np.allclose(h_dataset[()][...,1],radec.data.lat.value)
    assert h_dataset.attrs['lon_unit'] in ( radec.data.lon.unit.to_string().encode('ascii'),radec.data.lon.unit.to_string())
    assert h_dataset.attrs['lat_unit'] in ( radec.data.lat.unit.to_string().encode('ascii'),radec.data.lat.unit.to_string())
    reloaded = load_astropy.load_astropy_skycoord_dataset(h_dataset,b'astropy_skycoord',radec.__class__)
    assert np.allclose(reloaded.ra.value,radec.ra.value)
    assert np.allclose(reloaded.dec.value,radec.dec.value)

    ra = Angle([['1d20m', '0d21m'], ['1d20m', '0d21m']], unit='hourangle')
    dec = Angle([['33d0m0s', '33d01m'], ['33d0m0s', '33d01m']], unit='degree')
    radec = SkyCoord(ra, dec)
    h_dataset,subitems = load_astropy.create_astropy_skycoord(radec,h5_data,"astropy_skycoord_2",**compression_kwargs)
    assert isinstance(h_dataset,h5.Dataset) and not subitems and iter(subitems)
    assert np.allclose(h_dataset[()][...,0],radec.data.lon.value)
    assert np.allclose(h_dataset[()][...,1],radec.data.lat.value)
    assert h_dataset.attrs['lon_unit'] in ( radec.data.lon.unit.to_string().encode('ascii'),radec.data.lon.unit.to_string())
    assert h_dataset.attrs['lat_unit'] in ( radec.data.lat.unit.to_string().encode('ascii'),radec.data.lat.unit.to_string())
    reloaded = load_astropy.load_astropy_skycoord_dataset(h_dataset,b'astropy_skycoord',radec.__class__)
    assert np.allclose(reloaded.ra.value,radec.ra.value)
    assert np.allclose(reloaded.dec.value,radec.dec.value)
    assert reloaded.ra.shape == radec.ra.shape
    assert reloaded.dec.shape == radec.dec.shape

# %% MAIN SCRIPT
if __name__ == "__main__":
    from _pytest.fixtures import FixtureRequest
    from conftest import compression_kwargs
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_astropy_quantity),)
    ):
        test_create_astropy_quantity(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_astropy_constant),)
    ):
        test_create_astropy_constant(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_table),)
    ):
        test_astropy_table(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_quantity_array),)
    ):
        test_astropy_quantity_array(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_time_array),)
    ):
        test_astropy_time_array(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_angle),)
    ):
        test_astropy_angle(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_angle_array),)
    ):
        test_astropy_angle_array(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_skycoord),)
    ):
        test_astropy_skycoord(h5_root,keywords)
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_astropy_skycoord_array),)
    ):
        test_astropy_skycoord_array(h5_root,keywords)



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle.py

Unit test for hickle package.

"""


# %% IMPORTS

# Built-in imports
from collections import OrderedDict as odict
import os
import re
from pprint import pprint
import pickle


# Package imports
import numpy as np
from py.path import local
import pytest

# hickle imports
from hickle import dump, hickle, load, lookup

# Set current working directory to the temporary directory
local.get_temproot().chdir()


# %% GLOBALS
NESTED_DICT = {
    "level1_1": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_2": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_3": {
        "level2_1": {
            "level3_1": [1, 2, 3],
            "level3_2": [4, 5, 6]
        },
        "level2_2": [4, 5, 6]
    }
}


# %% FIXTURES
    
@pytest.fixture
def test_file_name(request):
    """
    create test dependent filename path string
    """
    yield "{}.hkl".format(request.function.__name__)


# %% HELPER DEFINITIONS

# Define a test function that must be serialized and unpacked again
def func(a, b, c=0):
    """ just something to do """
    return(a, b, c)

# the following is required as package name of with_state is hickle
# and load_loader refuses load any loader module for classes defined inside
# hickle package exempt when defined within load_*.py loaders modules.
# That has to be done by hickle sub modules directly using register_class function
pickle_dumps = pickle.dumps
pickle_loads = pickle.loads

types_to_hide = set() 

def make_visible_to_dumps(obj,protocol=None,*,fix_imports=True):
    """
    simulate loader functions defined outside hickle package
    """
    if obj in types_to_hide:
        obj.__module__ = re.sub(r'^\s*(?!hickle\.)','hickle.',obj.__module__)
    elif obj.__class__ in types_to_hide:
        obj.__class__.__module__ = re.sub(r'^\s*(?!hickle\.)','hickle.',obj.__class__.__module__)
    return pickle_dumps(obj,protocol,fix_imports=fix_imports)

def hide_from_hickle(bytes_obj,*,fix_imports=True,encoding="ASCII",errors="strict"):
    """
    simulate loader function defined outside hickle package
    """
    obj = pickle_loads(bytes_obj,fix_imports = fix_imports, encoding = encoding, errors = errors)
    if obj in types_to_hide:
        obj.__module__ = re.sub(r'^\s*hickle\.','',obj.__module__)
    elif obj.__class__ in types_to_hide:
        obj.__class__.__module__ = re.sub(r'^\s*hickle\.','',obj.__class__.__module__)
    return obj

# Define a class that must always be pickled
class with_state(object):
    """
    A class that always must be handled by create_pickled_dataset
    """
    def __init__(self):
        self.a = 12
        self.b = {
            'love': np.ones([12, 7]),
            'hatred': np.zeros([4, 9])}

    def __getstate__(self):
        self.a *= 2
        return({
            'a': self.a,
            'b': self.b})

    def __setstate__(self, state):
        self.a = state['a']
        self.b = state['b']

    def __getitem__(self, index):
        if(index == 0):
            return(self.a)
        if(index < 2):
            return(self.b['hatred'])
        if(index > 2):
            raise ValueError("index unknown")
        return(self.b['love'])

types_to_hide.add(with_state)

# %% FUNCTION DEFINITIONS
def test_invalid_file():
    """ Test if trying to use a non-file object fails. """

    with pytest.raises(hickle.FileError):
        dump('test', ())


def test_state_obj(monkeypatch,test_file_name,compression_kwargs):
    """ Dumping and loading a class object with pickle states

    https://github.com/telegraphic/hickle/issues/125"""

    with monkeypatch.context() as monkey:
        monkey.setattr(with_state,'__module__',re.sub(r'^\s*hickle\.','',with_state.__module__))
        monkey.setattr(pickle,'dumps',make_visible_to_dumps)
        mode = 'w'
        obj = with_state()
        with pytest.warns(lookup.SerializedWarning):
            dump(obj, test_file_name, mode,**compression_kwargs)
        monkey.setattr(pickle,'loads',hide_from_hickle)
        obj_hkl = load(test_file_name)
        assert isinstance(obj,obj_hkl.__class__) or isinstance(obj_hkl,obj.__class__)
        assert np.allclose(obj[1], obj_hkl[1])


def test_local_func(test_file_name,compression_kwargs):
    """ Dumping and loading a local function

    https://github.com/telegraphic/hickle/issues/119"""

    mode =  'w'
    with pytest.warns(lookup.SerializedWarning):
        dump(func, test_file_name, mode,**compression_kwargs)
    func_hkl = load(test_file_name)
    assert isinstance(func,func_hkl.__class__) or isinstance(func_hkl,func.__class__)
    assert func(1, 2) == func_hkl(1, 2)


def test_non_empty_group(test_file_name,compression_kwargs):
    """ Test if attempting to dump to a group with data fails """

    hickle.dump(None, test_file_name,**compression_kwargs)
    with pytest.raises(ValueError):
        dump(None, test_file_name, 'r+',**compression_kwargs)


def test_string(test_file_name,compression_kwargs):
    """ Dumping and loading a string """
    mode = 'w'
    string_obj = "The quick brown fox jumps over the lazy dog"
    dump(string_obj, test_file_name, mode,**compression_kwargs)
    string_hkl = load(test_file_name)
    assert isinstance(string_hkl, str)
    assert string_obj == string_hkl


def test_65bit_int(test_file_name,compression_kwargs):
    """ Dumping and loading an integer with arbitrary precision

    https://github.com/telegraphic/hickle/issues/113"""
    i = 2**65-1
    dump(i, test_file_name,**compression_kwargs)
    i_hkl = load(test_file_name)
    assert i == i_hkl

    j = -2**63-1
    dump(j, test_file_name,**compression_kwargs)
    j_hkl = load(test_file_name)
    assert j == j_hkl

def test_list(test_file_name,compression_kwargs):
    """ Dumping and loading a list """
    filename, mode = 'test_list.h5', 'w'
    list_obj = [1, 2, 3, 4, 5]
    dump(list_obj, test_file_name, mode=mode,**compression_kwargs)
    list_hkl = load(test_file_name)
    try:
        assert isinstance(list_hkl, list)
        assert list_obj == list_hkl
        import h5py
        a = h5py.File(test_file_name, 'r')
        a.close()

    except AssertionError:
        print("ERR:", list_obj, list_hkl)
        import h5py

        raise


def test_set(test_file_name,compression_kwargs)    :
    """ Dumping and loading a list """
    mode = 'w'
    list_obj = set([1, 0, 3, 4.5, 11.2])
    dump(list_obj, test_file_name, mode,**compression_kwargs)
    list_hkl = load(test_file_name)
    try:
        assert isinstance(list_hkl, set)
        assert list_obj == list_hkl
    except AssertionError:
        print(type(list_obj))
        print(type(list_hkl))
        raise


def test_numpy(test_file_name,compression_kwargs):
    """ Dumping and loading numpy array """
    mode = 'w'
    dtypes = ['float32', 'float64', 'complex64', 'complex128']

    for dt in dtypes:
        array_obj = np.ones(8, dtype=dt)
        dump(array_obj, test_file_name, mode,**compression_kwargs)
        array_hkl = load(test_file_name)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
    except AssertionError:
        print(array_hkl)
        print(array_obj)
        raise


def test_masked(test_file_name,compression_kwargs):
    """ Test masked numpy array """
    mode = 'w'
    a = np.ma.array([1, 2, 3, 4], dtype='float32', mask=[0, 1, 0, 0])

    dump(a, test_file_name, mode,**compression_kwargs)
    a_hkl = load(test_file_name)

    try:
        assert a_hkl.dtype == a.dtype
        assert np.all((a_hkl, a))
    except AssertionError:
        print(a_hkl)
        print(a)
        raise


def test_object_numpy(test_file_name,compression_kwargs):
    """ Dumping and loading a NumPy array containing non-NumPy objects.

    https://github.com/telegraphic/hickle/issues/90"""

    # VisibleDeprecationWarning from newer numpy versions
    #np_array_data = np.array([[NESTED_DICT], ('What is this?',), {1, 2, 3, 7, 1}])
    arr = np.array([NESTED_DICT])#, ('What is this?',), {1, 2, 3, 7, 1}])
    dump(arr, test_file_name,**compression_kwargs)
    arr_hkl = load(test_file_name)
    assert np.all(arr == arr_hkl)

    arr2 = np.array(NESTED_DICT)
    dump(arr2, test_file_name,**compression_kwargs)
    arr_hkl2 = load(test_file_name)
    assert np.all(arr2 == arr_hkl2)


def test_string_numpy(test_file_name,compression_kwargs):
    """ Dumping and loading NumPy arrays containing Python 3 strings. """

    arr = np.array(["1313e", "was", "maybe?", "here"])
    dump(arr, test_file_name,**compression_kwargs)
    arr_hkl = load(test_file_name)
    assert np.all(arr == arr_hkl)


def test_list_object_numpy(test_file_name,compression_kwargs):
    """ Dumping and loading a list of NumPy arrays with objects.

    https://github.com/telegraphic/hickle/issues/90"""

    # VisibleDeprecationWarning from newer numpy versions
    lst = [np.array(NESTED_DICT)]#, np.array([('What is this?',),
                                 #           {1, 2, 3, 7, 1}])]
    dump(lst, test_file_name,**compression_kwargs)
    lst_hkl = load(test_file_name)
    assert np.all(lst[0] == lst_hkl[0])
    #assert np.all(lst[1] == lst_hkl[1])


def test_dict(test_file_name,compression_kwargs):
    """ Test dictionary dumping and loading """
    mode = 'w'

    dd = {
        'name': b'Danny',
        'age': 28,
        'height': 6.1,
        'dork': True,
        'nums': [1, 2, 3],
        'narr': np.array([1, 2, 3]),
    }

    dump(dd, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()

            if isinstance(dd[k], np.ndarray):
                assert np.all((dd[k], dd_hkl[k]))
            else:
                pass
            assert isinstance(dd_hkl[k], dd[k].__class__)
        except AssertionError:
            print(k)
            print(dd_hkl[k])
            print(dd[k])
            print(type(dd_hkl[k]), type(dd[k]))
            raise


def test_odict(test_file_name,compression_kwargs):
    """ Test ordered dictionary dumping and loading

    https://github.com/telegraphic/hickle/issues/65"""
    mode = 'w'

    od = odict(((3, [3, 0.1]), (7, [5, 0.1]), (5, [3, 0.1])))
    dump(od, test_file_name, mode,**compression_kwargs)
    od_hkl = load(test_file_name)

    assert od.keys() == od_hkl.keys()

    for od_item, od_hkl_item in zip(od.items(), od_hkl.items()):
        assert od_item == od_hkl_item


def test_empty_dict(test_file_name,compression_kwargs):
    """ Test empty dictionary dumping and loading

    https://github.com/telegraphic/hickle/issues/91"""
    mode = 'w'

    dump({}, test_file_name, mode,**compression_kwargs)
    assert load(test_file_name) == {}



# TODO consider converting to parameterized test
#      or enable implicit parameterizing of all tests
#      though compression_kwargs fixture providing
#      various combinations of compression and chunking
#      related keywords
@pytest.mark.no_compression
def test_compression(test_file_name):
    """ Test compression on datasets"""

    mode = 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']

    comps = [None, 'gzip', 'lzf']

    for dt in dtypes:
        for cc in comps:
            array_obj = np.ones(32768, dtype=dt)
            dump(array_obj, test_file_name, mode, compression=cc)
            print(cc, os.path.getsize(test_file_name))
            array_hkl = load(test_file_name)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
    except AssertionError:
        print(array_hkl)
        print(array_obj)
        raise


def test_dict_int_key(test_file_name,compression_kwargs):
    """ Test for dictionaries with integer keys """
    mode = 'w'

    dd = {
        0: "test",
        1: "test2"
    }

    dump(dd, test_file_name, mode,**compression_kwargs)
    load(test_file_name)


def test_dict_nested(test_file_name,compression_kwargs):
    """ Test for dictionaries with integer keys """
    mode = 'w'

    dd = NESTED_DICT

    dump(dd, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    ll_hkl = dd_hkl["level1_3"]["level2_1"]["level3_1"]
    ll = dd["level1_3"]["level2_1"]["level3_1"]
    assert ll == ll_hkl


def test_masked_dict(test_file_name,compression_kwargs):
    """ Test dictionaries with masked arrays """

    filename, mode = 'test.h5', 'w'

    dd = {
        "data": np.ma.array([1, 2, 3], mask=[True, False, False]),
        "data2": np.array([1, 2, 3, 4, 5])
    }

    dump(dd, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()
            if isinstance(dd[k], np.ndarray):
                assert np.all((dd[k], dd_hkl[k]))
            elif isinstance(dd[k], np.ma.MaskedArray):
                print(dd[k].data)
                print(dd_hkl[k].data)
                assert np.allclose(dd[k].data, dd_hkl[k].data)
                assert np.allclose(dd[k].mask, dd_hkl[k].mask)

            assert isinstance(dd_hkl[k], dd[k].__class__)

        except AssertionError:
            print(k)
            print(dd_hkl[k])
            print(dd[k])
            print(type(dd_hkl[k]), type(dd[k]))
            raise


def test_np_float(test_file_name,compression_kwargs):
    """ Test for singular np dtypes """
    mode = 'w'

    dtype_list = (np.float16, np.float32, np.float64,
                  np.complex64, np.complex128,
                  np.int8, np.int16, np.int32, np.int64,
                  np.uint8, np.uint16, np.uint32, np.uint64)

    for dt in dtype_list:

        dd = dt(1)
        dump(dd, test_file_name, mode,**compression_kwargs)
        dd_hkl = load(test_file_name)
        assert dd == dd_hkl
        assert dd.dtype == dd_hkl.dtype

    dd = {}
    for dt in dtype_list:
        dd[str(dt)] = dt(1.0)
    dump(dd, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    print(dd)
    for dt in dtype_list:
        assert dd[str(dt)] == dd_hkl[str(dt)]


# TODO consider converting to parameterized test
#      or enable implicit parameterizing of all tests
#      though compression_kwargs fixture providing
#      various combinations of compression and chunking
#      related keywords
@pytest.mark.no_compression
def test_comp_kwargs(test_file_name):
    """ Test compression with some kwargs for shuffle and chunking """

    mode = 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']

    comps = [None, 'gzip', 'lzf']
    chunks = [(100, 100), (250, 250)]
    shuffles = [True, False]
    scaleoffsets = [0, 1, 2]

    for dt in dtypes:
        for cc in comps:
            for ch in chunks:
                for sh in shuffles:
                    for so in scaleoffsets:
                        kwargs = {
                            'compression': cc,
                            'dtype': dt,
                            'chunks': ch,
                            'shuffle': sh,
                            'scaleoffset': so
                        }
                        array_obj = NESTED_DICT
                        dump(array_obj, test_file_name, mode, compression=cc)
                        print(kwargs, os.path.getsize(test_file_name))
                        load(test_file_name)


def test_list_numpy(test_file_name,compression_kwargs):
    """ Test converting a list of numpy arrays """

    mode = 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = [a, b]

    dump(c, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    print(dd_hkl)

    assert isinstance(dd_hkl, list)
    assert isinstance(dd_hkl[0], np.ndarray)


def test_tuple_numpy(test_file_name,compression_kwargs):
    """ Test converting a list of numpy arrays """

    mode = 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = (a, b, a)

    dump(c, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)

    print(dd_hkl)

    assert isinstance(dd_hkl, tuple)
    assert isinstance(dd_hkl[0], np.ndarray)


def test_numpy_dtype(test_file_name,compression_kwargs):
    """ Dumping and loading a NumPy dtype """

    dtype = np.dtype('int64')
    dump(dtype, test_file_name,**compression_kwargs)
    dtype_hkl = load(test_file_name)
    assert dtype == dtype_hkl


def test_none(test_file_name,compression_kwargs):
    """ Test None type hickling """

    mode = 'w'

    a = None

    dump(a, test_file_name, mode,**compression_kwargs)
    dd_hkl = load(test_file_name)
    print(a)
    print(dd_hkl)

    assert isinstance(dd_hkl, type(None))


def test_list_order(test_file_name,compression_kwargs):
    """ https://github.com/telegraphic/hickle/issues/26 """
    d = [np.arange(n + 1) for n in range(20)]
    dump(d, test_file_name,**compression_kwargs)
    d_hkl = load(test_file_name)

    try:
        for ii, xx in enumerate(d):
            assert d[ii].shape == d_hkl[ii].shape
        for ii, xx in enumerate(d):
            assert np.allclose(d[ii], d_hkl[ii])
    except AssertionError:
        print(d[ii], d_hkl[ii])
        raise


def test_embedded_array(test_file_name,compression_kwargs):
    """ See https://github.com/telegraphic/hickle/issues/24 """

    d_orig = [[np.array([10., 20.]), np.array([10, 20, 30])],
              [np.array([10, 2]), np.array([1.])]]
    dump(d_orig, test_file_name,**compression_kwargs)
    d_hkl = load(test_file_name)

    for ii, xx in enumerate(d_orig):
        for jj, yy in enumerate(xx):
            assert np.allclose(d_orig[ii][jj], d_hkl[ii][jj])

    print(d_hkl)
    print(d_orig)


##############
# NEW TESTS  #
###############
def generate_nested():
    a = [1, 2, 3]
    b = [a, a, a]
    c = [a, b, 's']
    d = [a, b, c, c, a]
    e = [d, d, d, d, 1]
    f = {'a': a, 'b': b, 'e': e}
    g = {'f': f, 'a': e, 'd': d}
    h = {'h': g, 'g': f}
    z = [f, a, b, c, d, e, f, g, h, g, h]
    a = np.array([1, 2, 3, 4])
    b = set([1, 2, 3, 4, 5])
    c = (1, 2, 3, 4, 5)
    d = np.ma.array([1, 2, 3, 4, 5, 6, 7, 8])
    z = {'a': a, 'b': b, 'c': c, 'd': d, 'z': z}
    return z

def test_dump_nested(test_file_name,compression_kwargs):
    """ Dump a complicated nested object to HDF5
    """
    z = generate_nested()
    dump(z, test_file_name, mode='w',**compression_kwargs)

def test_ndarray(test_file_name,compression_kwargs):
    a = np.array([1, 2, 3])
    b = np.array([2, 3, 4])
    z = (a, b)

    print("Original:")
    pprint(z)
    dump(z, test_file_name, mode='w',**compression_kwargs)

    print("\nReconstructed:")
    z = load(test_file_name)
    pprint(z)


def test_ndarray_masked(test_file_name,compression_kwargs):
    a = np.ma.array([1, 2, 3])
    b = np.ma.array([2, 3, 4], mask=[True, False, True])
    z = (a, b)

    print("Original:")
    pprint(z)
    dump(z, test_file_name, mode='w',**compression_kwargs)

    print("\nReconstructed:")
    z = load(test_file_name)
    pprint(z)


def test_simple_dict(test_file_name,compression_kwargs):
    a = {'key1': 1, 'key2': 2}

    dump(a, test_file_name,**compression_kwargs)
    z = load(test_file_name)

    pprint(a)
    pprint(z)


def test_complex_dict(test_file_name,compression_kwargs):
    a = {'akey': 1, 'akey2': 2}
    c = {'ckey': "hello", "ckey2": "hi there"}
    z = {'zkey1': a, 'zkey2': a, 'zkey3': c}

    print("Original:")
    pprint(z)
    dump(z, test_file_name, mode='w',**compression_kwargs)

    print("\nReconstructed:")
    z = load(test_file_name)
    pprint(z)

def test_complex(test_file_name,compression_kwargs):
    """ Test complex value dtype is handled correctly

    https://github.com/telegraphic/hickle/issues/29 """

    data = {"A": 1.5, "B": 1.5 + 1j, "C": np.linspace(0, 1, 4) + 2j}
    dump(data, test_file_name,**compression_kwargs)
    data2 = load(test_file_name)
    for key in data.keys():
        assert isinstance(data[key], data2[key].__class__)


def test_nonstring_keys(test_file_name,compression_kwargs):
    """ Test that keys are reconstructed back to their original datatypes
    https://github.com/telegraphic/hickle/issues/36
    """

    data = {
            u'test': 123,
            'def': [b'test'],
            'hik': np.array([1, 2, 3]),
            0: 0,
            True: ['test'],
            1.1: 'hey',
            1j: 'complex_hashable',
            (1, 2): 'boo',
            ('A', 17.4, 42): [1, 7, 'A'],
            (): '1313e was here',
            '0': 0,
            None: None
            }

    print(data)
    dump(data, test_file_name,**compression_kwargs)
    data2 = load(test_file_name)
    print(data2)

    for key in data.keys():
        assert key in data2.keys()

    print(data2)

@pytest.mark.no_compression
def test_scalar_compression(test_file_name):
    """ Test bug where compression causes a crash on scalar datasets

    (Scalars are incompressible!)
    https://github.com/telegraphic/hickle/issues/37
    """
    data = {'a': 0, 'b': np.float(2), 'c': True}

    dump(data, test_file_name, compression='gzip')
    data2 = load(test_file_name)

    print(data2)
    for key in data.keys():
        assert isinstance(data[key], data2[key].__class__)


def test_bytes(test_file_name,compression_kwargs):
    """ Dumping and loading a string. PYTHON3 ONLY """

    mode = 'w'
    string_obj = b"The quick brown fox jumps over the lazy dog"
    dump(string_obj, test_file_name, mode,**compression_kwargs)
    string_hkl = load(test_file_name)
    print(type(string_obj))
    print(type(string_hkl))
    assert isinstance(string_hkl, bytes)
    assert string_obj == string_hkl


def test_np_scalar(test_file_name,compression_kwargs):
    """ Numpy scalar datatype

    https://github.com/telegraphic/hickle/issues/50
    """

    r0 = {'test': np.float64(10.)}
    dump(r0, test_file_name,**compression_kwargs)
    r = load(test_file_name)
    print(r)
    assert isinstance(r0['test'], r['test'].__class__)


def test_slash_dict_keys(test_file_name,compression_kwargs):
    """ Support for having slashes in dict keys

    https://github.com/telegraphic/hickle/issues/124"""
    dct = {'a/b': [1, '2'], 1.4: 3}

    dump(dct, test_file_name, 'w',**compression_kwargs)
    dct_hkl = load(test_file_name)

    assert isinstance(dct_hkl, dict)
    for key, val in dct_hkl.items():
        assert val == dct.get(key)

    # Check that having backslashes in dict keys will serialize the dict
    dct2 = {'a\\b': [1, '2'], 1.4: 3}
    with pytest.warns(None) as not_expected:
        dump(dct2, test_file_name,**compression_kwargs)
    assert not not_expected


# %% MAIN SCRIPT
if __name__ == '__main__':
    """ Some tests and examples """
    from _pytest.fixtures import FixtureRequest

    for filename in test_file_name(FixtureRequest(test_np_scalar)):
        test_np_scalar(filename)
    for filename in test_file_name(FixtureRequest(test_scalar_compression)):
        test_scalar_compression(filename)
    for filename in test_file_name(FixtureRequest(test_complex)):
        test_complex(filename)
    for filename in test_file_name(FixtureRequest(test_none)):
        test_none(filename)
    for filename in test_file_name(FixtureRequest(test_masked_dict)):
        test_masked_dict(filename)
    for filename in test_file_name(FixtureRequest(test_list)):
        test_list(filename)
    for filename in test_file_name(FixtureRequest(test_set)):
        test_set(filename)
    for filename in test_file_name(FixtureRequest(test_numpy)):
        test_numpy(filename)
    for filename in test_file_name(FixtureRequest(test_dict)):
        test_dict(filename)
    for filename in test_file_name(FixtureRequest(test_odict)):
        test_odict(filename)
    for filename in test_file_name(FixtureRequest(test_empty_dict)):
        test_empty_dict(filename)
    for filename in test_file_name(FixtureRequest(test_compression)):
        test_compression(filename)
    for filename in test_file_name(FixtureRequest(test_masked)):
        test_masked(filename)
    for filename in test_file_name(FixtureRequest(test_dict_nested)):
        test_dict_nested(filename)
    for filename in test_file_name(FixtureRequest(test_comp_kwargs)):
        test_comp_kwargs(filename)
    for filename in test_file_name(FixtureRequest(test_list_numpy)):
        test_list_numpy(filename)
    for filename in test_file_name(FixtureRequest(test_tuple_numpy)):
        test_tuple_numpy(filename)
    for filename in test_file_name(FixtureRequest(test_list_order)):
        test_list_order(filename)
    for filename in test_file_name(FixtureRequest(test_embedded_array)):
        test_embedded_array(filename)
    for filename in test_file_name(FixtureRequest(test_np_float)):
        test_np_float(filename)
    for filename in test_file_name(FixtureRequest(test_string)):
        test_string(filename)
    for filename in test_file_name(FixtureRequest(test_nonstring_keys)):
        test_nonstring_keys(filename)
    for filename in test_file_name(FixtureRequest(test_bytes)):
        test_bytes(filename)

    # NEW TESTS
    for filename in test_file_name(FixtureRequest(test_dump_nested)):
        test_dump_nested(filename)
    for filename in test_file_name(FixtureRequest(test_ndarray)):
        test_ndarray(filename)
    for filename in test_file_name(FixtureRequest(test_ndarray_masked)):
        test_ndarray_masked(filename)
    for filename in test_file_name(FixtureRequest(test_simple_dict)):
        test_simple_dict(filename)
    for filename in test_file_name(FixtureRequest(test_complex_dict)):
        test_complex_dict(filename)
    for filename in test_file_name(FixtureRequest(test_dict_int_key)):
        test_dict_int_key(filename)
    for filename in test_file_name(FixtureRequest(test_local_func)):
        test_local_func(filename)
    for filename in test_file_name(FixtureRequest(test_slash_dict_keys)):
        test_slash_dict_keys(filename)
    test_invalid_file()
    for filename in test_file_name(FixtureRequest(test_non_empty_group)):
        test_non_empty_group(filename)
    for filename in test_file_name(FixtureRequest(test_numpy_dtype)):
        test_numpy_dtype(filename)
    for filename in test_file_name(FixtureRequest(test_object_numpy)):
        test_object_numpy(filename)
    for filename in test_file_name(FixtureRequest(test_string_numpy)):
        test_string_numpy(filename)
    for filename in test_file_name(FixtureRequest(test_list_object_numpy)):
        test_list_object_numpy(filename)

    # Cleanup
    for filename in test_file_name(FixtureRequest(print)):
        print(filename)



# %% IMPORTS
# Built-in imports
import glob
from os import path
import warnings
import pytest
import scipy.sparse
import numpy as np

# Package imports
import h5py

# hickle imports
import hickle as hkl


# %% FUNCTION DEFINITIONS
def test_legacy_load():
    dirpath = path.dirname(__file__)
    filelist = sorted(glob.glob(path.join(dirpath, 'legacy_hkls/*3_[0-9]_[0-9].hkl')))

    # Make all warnings show
    warnings.simplefilter("always")

    for filename in filelist:
        with pytest.warns(
            UserWarning,
            match = r"Input\s+argument\s+'file_obj'\s+appears\s+to\s+be\s+a\s+file\s+made"
                    r"\s+with\s+hickle\s+v3.\s+Using\s+legacy\s+load..."
        ):
            try:
                print(filename)
                a = hkl.load(filename,path='test')
            except Exception:
                with h5py.File(filename) as a:
                    print(a.attrs.items())
                    print(a.items())
                    for key, item in a.items():
                        print(item.attrs.items())
                    raise

@pytest.mark.no_compression
def test_4_0_0_load():
    """
    test that files created by hickle 4.0.x can be loaded by 
    hickle 4.1.x properly
    """
    dirpath = path.dirname(__file__)
    filelist = sorted(glob.glob(path.join(dirpath, 'legacy_hkls/*4.[0-9].[0-9].hkl')))
    from hickle.tests.generate_legacy_4_0_0 import generate_py_object
    compare_with,needs_compare = generate_py_object()
    # strange but without forcing garbage collection here h5py might produce 
    # strange assuming a race related RuntimeError when h5py file is closed by
    # hickle.load(). Unless observed in wildlife this is only triggered by fast successive
    # calls of h5py methods.
    import gc
    gc.collect()
    for filename in filelist:
        content = hkl.load(filename)
        if filename != needs_compare:
            continue
        for item_id,content_item,compare_item in ( (i,content[i],compare_with[i]) for i in range(len(compare_with)) ):
            if scipy.sparse.issparse(content_item):
                assert np.allclose(content_item.toarray(),compare_item.toarray())
                continue
            try:
                assert content_item == compare_item
            except ValueError:
                assert np.all(content_item == compare_item)

# %% MAIN SCRIPT
if __name__ == "__main__":
    test_legacy_load()
    test_4_0_0_load()



# To increment version
# Check you have ~/.pypirc filled in
# git tag x.y.z
# git push && git push --tags
# rm -rf dist; python setup.py sdist bdist_wheel
# TEST: twine upload --repository-url https://test.pypi.org/legacy/ dist/*
# twine upload dist/*

from codecs import open
import re

from setuptools import setup, find_packages
import sys

author = "Danny Price, Ellert van der Velden and contributors"

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", 'r') as fh:
    requirements = fh.read().splitlines()

with open("requirements_test.txt", 'r') as fh:
    test_requirements = fh.read().splitlines()

# Read the __version__.py file
with open('hickle/__version__.py', 'r') as f:
    vf = f.read()

# Obtain version from read-in __version__.py file
version = re.search(r"^_*version_* = ['\"]([^'\"]*)['\"]", vf, re.M).group(1)

setup(name='hickle',
      version=version,
      description='Hickle - an HDF5 based version of pickle',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author=author,
      author_email='dan@thetelegraphic.com',
      url='http://github.com/telegraphic/hickle',
      download_url=('https://github.com/telegraphic/hickle/archive/v%s.zip'
                    % (version)),
      platforms='Cross platform (Linux, Mac OSX, Windows)',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Utilities',
          ],
      keywords=['pickle', 'hdf5', 'data storage', 'data export'],
      install_requires=requirements,
      tests_require=test_requirements,
      python_requires='>=3.5',
      packages=find_packages(),
      zip_safe=False,
)



# %% IMPORTS
# Package imports
import pickle
import numpy as np
import scipy
import copy
from scipy import sparse

# hickle imports
from hickle.helpers import PyContainer,H5NodeFilterProxy
from hickle.loaders.load_numpy import load_ndarray_dataset,create_np_array_dataset


# %% FUNCTION DEFINITIONS
def return_first(x):
    """
    Dummy function used as place holder type in loading legacy hickle 4.x files
    """
    raise TypeError("'return_first' not callable and deprecated. Create and use PyContainer instead.")


def create_sparse_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an sparse array to h5py file

    Parameters
    ----------
    py_obj (scipy.sparse.csr_matrix,scipy.sparse.csc_matrix, scipy.sparse.bsr_matrix):
        python object to dump

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns:
        Group and list of subitems to dump into
    """
    h_sparsegroup = h_group.create_group(name)
    return h_sparsegroup,(
        ('data',py_obj.data,{},kwargs),
        ('indices',py_obj.indices,{},kwargs),
        ('indptr',py_obj.indptr,{},kwargs),
        ('shape',py_obj.shape,{},kwargs)
    )

ndarray_type_string = pickle.dumps(np.ndarray)
tuple_type_string = pickle.dumps(tuple)

            
class SparseMatrixContainer(PyContainer):
    """
    PyContainer used to restore sparse Matrix
    """

    # instance attribute shadowing class method of same name
    # points per default to shadowed method
    __slots__ = ('filter',)
    
    _index_name_map = {
        'data':0,
        'indices':1,
        'indptr':2,
        'shape':3
    }

    def __init__(self,h5_attrs, base_type, object_type):
        super(SparseMatrixContainer,self).__init__(h5_attrs,base_type,object_type,_content = [None]*4)
        
        # in case object type is return_first (hickle 4.x file) than switch filter
        # to redirect loading of sub items to numpy ndarray type. Otherwise set to 
        # PyContainer.filter method
        if object_type is return_first:
            self.filter = self._redirect_to_ndarray
        else:
            self.filter = super(SparseMatrixContainer,self).filter

    def _redirect_to_ndarray(self,h_parent):
        """
        iterates through items and extracts effective object and basetype
        of sparse matrix from data subitem and remaps all subitems to 
        ndarray type exempt shape which is remapped to tuple
        """

        for name,item in h_parent.items():
            item = H5NodeFilterProxy(item)
            if name == 'data':
                self.object_type = pickle.loads(item.attrs['type'])
                self.base_type = item.attrs['base_type']
                np_dtype = item.attrs.get('np_dtype',None)
                if np_dtype is None:
                    item.attrs['np_dtype'] = item.dtype.str.encode('ascii')
            elif name not in self._index_name_map.keys():
                continue # ignore name
            if name == "shape":
                item.attrs['type'] = np.array(tuple_type_string)
                item.attrs['base_type'] = b'tuple'
            else:
                item.attrs['type'] = np.array(ndarray_type_string)
                item.attrs['base_type'] = b'ndarray'
                np_dtype = item.attrs.get('np_dtype',None)
                if np_dtype is None:
                    item.attrs['np_dtype'] = item.dtype.str.encode('ascii')
            yield name,item

    def append(self,name,item,h5_attrs):
        index = self._index_name_map.get(name,None)
        self._content[index] = item

    def convert(self):
        return self.object_type(tuple(self._content[:3]),dtype=self._content[0].dtype,shape=self._content[3])

# %% REGISTERS
class_register = [
    [scipy.sparse.csr_matrix, b'csr_matrix', create_sparse_dataset, None, SparseMatrixContainer],
    [scipy.sparse.csc_matrix, b'csc_matrix', create_sparse_dataset, None, SparseMatrixContainer],
    [scipy.sparse.bsr_matrix, b'bsr_matrix', create_sparse_dataset, None, SparseMatrixContainer]
]

exclude_register = []




def create_package_test(myclass_type,h_group,name,**kwargs):
    return h_group,()

def load_package_test(h_node,base_type,py_obj_type):
    return {12:12}


class_register = [
    ( dict,b'dict',create_package_test,load_package_test )
]

exclude_register = [b'please_kindly_ignore_me']
